/*
 * Copyright 1995 Steve Passe
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of AUTHOR (Steve Passe)
 * is not be used in advertising or publicity pertaining to distribution of
 * the software without specific, written prior permission.  AUTHOR makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL AUTHOR
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * AUTHOR:     Steve Passe <smp@csn.net>
 */


static char* rcsid = "$Id: vm100.c,v 1.3 1995/08/27 05:08:59 smp Exp smp $";

#include <stdio.h>
#include <errno.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>

#include "vm100.h"

#if defined( DEBUG )
FILE* debugLog = NULL;
#endif


#define TRUE			(1)
#define	FALSE			(0)

/* 1/4 second without ping */
#define RING_CYCLE_TIME		(250000L)

/* minimum of 5 pings */
#define MIN_RING_CYCLES		(10)


/** should be class member data */
int vm100Port;
int parent = 0;

/** should be class member data */
int ringState = FALSE;
int ringCycleCount = 0;

static int	vm100Run( char* port );
static int	vm100HardInitialize( char* port );
static int	vm100SoftInitialize( void );

static void	vm100OffHookAction( int sig, int code, struct sigcontext *scp);
static void	vm100OnHookAction( int sig, int code, struct sigcontext *scp);
static void	vm100RingStart( int sig, int code, struct sigcontext *scp );
static void	vm100RingStop( int sig, int code, struct sigcontext *scp );


/*****************************************************************************
 * public functions
 */

/*
 * song and dance time, fork a child & use setsid() to create
 * our own session. this allows us to make the vm100 port the
 * controlling port, which gets us SIGHUP signals for ring detect [assumming
 * that we have strapped vm100 pin 5 (CTS) to tty pin 8 (DCD)].
 */
int
vm100Daemon( char* port,
	     void (*ringStartCallback)(),
	     void (*ringStopCallback)() )
{
    struct sigaction action;
    int forkStatus;

    /* fork failed */
    if ( (forkStatus = fork()) < 0 )
    {
	perror( "vm100: child fork" );
	exit( 1 );
    }

    /* we're the parent */
    else if ( forkStatus > 0 )
    {
	action.sa_flags = SA_RESTART;

	action.sa_handler = ringStartCallback ? ringStartCallback : SIG_IGN;
	if ( sigaction( SIGUSR1, &action, NULL ) < 0 )
	{
	    return BAD;
	}

	action.sa_handler = ringStopCallback ? ringStopCallback : SIG_IGN;
	if ( sigaction( SIGUSR2, &action, NULL ) < 0 )
	{
	    return BAD;
	}

	/* return the process id of the vm100 daemon */
	return forkStatus;
    }

    /* we're the child */
    else
    {

#if defined( DEBUG )
	debugLog = fopen( "/usr/tmp/phonelog", "a" ); fflush(debugLog);
#endif

	/* give up stuff */
	close( 0 );
	close( 1 );
	close( 2 );

	/* become the process leader */
	if ( setsid() < 0 )
	{

#if defined( DEBUG )
	    fprintf( debugLog, "setsid()\n" );
#endif

	    exit( 1 );
	}

	vm100Run( port );

#if defined( DEBUG )
	fprintf( debugLog, "exit(0)\n" );
#endif

	exit( 0 );
    }
}


/*
 * take the phone OFF-HOOK, ie. pickup
 */
int
vm100OffHook( int daemon )
{
    if ( kill( daemon, SIGUSR1 ) < 0 )
    {

#if defined( DEBUG )
	fprintf( debugLog, "kill( SIGUSR1 )\n" );
#endif

	exit( 1 );
    }
}


/*
 * put the phone ON-HOOK, ie. hangup
 */
int
vm100OnHook( int daemon )
{
    if ( kill( daemon, SIGUSR2 ) < 0 )
    {

#if defined( DEBUG )
	fprintf( debugLog, "kill( SIGUSR2 )\n" );
#endif

	exit( 1 );
    }
}


/*****************************************************************************
 * private functions
 */

/*
 * 
 */
static int
vm100Run( char* port )
{
    int status;

    /* initialize the vm100 dongle */
    if ( vm100HardInitialize( port ) == BAD )
    {

#if defined( DEBUG )
	fprintf( debugLog, "hard initialize\n" );
#endif

	exit( 2 );
    }

    /*  */
    if ( vm100SoftInitialize() == BAD )
    {
#if defined( DEBUG )
	fprintf( debugLog, "soft initialize\n" );
#endif
	exit( 3 );
    }

    for ( ; /* ever */; )
    {
	/*
	 * signals do all the work:
	 *  receive SIGUSR1: go off-hook
	 *  receive SIGUSR2: go off-hook
	 *  ring detect: send SIGUSR1
	 */
	wait( &status );
    }

}


/*
 * 
 */
static int
vm100HardInitialize( char* port )
{
    int portState;

    /* open port */
    if ( (vm100Port = open( port, O_RDWR | O_NONBLOCK )) < 0 )
    {
	return BAD;
    }

    /* go on-hook ASAP, open() might take phone line off hook! */
    vm100OnHookAction( NULL, NULL, NULL );

    /* initialize DTR to provide power to device */
    portState = TIOCM_DTR;
    if ( ioctl( vm100Port, TIOCMBIS, &portState ) < 0 )
    {
	return BAD;
    }

    return OK;
}


static int
vm100SoftInitialize( void )
{
    struct sigaction action;

    /*  */
    action.sa_flags = SA_RESTART;

    /* catch START RING */
    action.sa_handler = vm100RingStart;
    if ( sigaction( SIGHUP, &action, NULL ) < 0 )
    {
	return BAD;
    }

    /* catch STOP RING */
    action.sa_handler = vm100RingStop;
    if ( sigaction( SIGALRM, &action, NULL ) < 0 )
    {
	return BAD;
    }

    /* catch SIGUSR1 */
    action.sa_handler = vm100OffHookAction;
    if ( sigaction( SIGUSR1, &action, NULL ) < 0 )
    {
	return BAD;
    }

    /* catch SIGUSR2 */
    action.sa_handler = vm100OnHookAction;
    if ( sigaction( SIGUSR2, &action, NULL ) < 0 )
    {
	return BAD;
    }

    /* arrange to get SIGHUPs from vm100's ring detect (CTS hooked to DCD) */
    if ( ioctl( vm100Port, TIOCSCTTY, NULL ) < 0 )
    {
	return BAD;
    }

    /* get pid of our parent for ring detect signals */
    parent = getppid();

    return OK;
}


/*
 * signal handler to take the phone OFF-HOOK, ie. pickup
 */
static void
vm100OffHookAction( int sig, int code, struct sigcontext *scp )
{
    const int portState = TIOCM_RTS;
    if ( ioctl( vm100Port, TIOCMBIS, &portState ) < 0 )
    {

#if defined( DEBUG )
	fprintf( debugLog, "OffHAction\n" );
#endif

	exit( 1 );
    }
}


/*
 * signal handler to put the phone ON-HOOK, ie. hangup
 */
static void
vm100OnHookAction( int sig, int code, struct sigcontext *scp )
{
    const int portState = TIOCM_RTS;
    if ( ioctl( vm100Port, TIOCMBIC, &portState ) < 0 )
    {

#if defined( DEBUG )
	fprintf( debugLog, "OnHAction\n" );
#endif

	exit( 1 );
    }
}


/*
 * ring detect signal received
 */
static void
vm100RingStart( int sig, int code, struct sigcontext *scp )
{
    const struct itimerval value = { { 0L, RING_CYCLE_TIME },
				     { 0L, RING_CYCLE_TIME } };

    /* look for first cycle */
    if ( ringState == FALSE )
    {

#if defined( DEBUG )
	putc( 'x', debugLog ); fflush( debugLog );
#endif

	/* filter noise */
	if ( ++ringCycleCount >= MIN_RING_CYCLES )
	{
	    ringState = TRUE;
	    if ( kill( parent, SIGUSR1 ) < 0 )
	    {

#if defined( DEBUG )
		fprintf( debugLog, "ringStart\n" );
#endif

		exit( 1 );
	    }
	}
    }

    /* cock timeout timer */
    setitimer( ITIMER_REAL, &value, NULL );
}


/*
 * timeout on ring detect,
 *  ie., no detect within last RING_CYCLE_TIME microseconds
 */
static void
vm100RingStop( int sig, int code, struct sigcontext *scp )
{
    const struct itimerval value = { { 0L, 0L }, { 0L, 0L } };

    if ( ringState == TRUE )
    {
	if ( kill( parent, SIGUSR2 ) < 0 )
	{

#if defined( DEBUG )
	    fprintf( debugLog, "ringStop\n" );
#endif

	    exit( 1 );
	}
	ringState = FALSE;
    }

    /* clear ring 'noise filter' */
    ringCycleCount = 0;

    /* clear timeout timer */
    setitimer( ITIMER_REAL, &value, NULL );
}
